//
//                                                                         
// Module: Internals Example Source File                                   
//                                                                         
// Description: Declarations for the Internals Example Plugin              
//                                                                         
//                                                                         
// This source code module, and all information, data, and algorithms      
// associated with it, are part of CUBE technology (tm).                   
//                           PROPRIETARY                                   
// Copyright (c) 1996-2006 Image Space Incorporated.  All rights reserved. 
//                                                                         
//                                                                         
// Change history:                                                         
//   tag.2005.11.30: created                                               
//   jmm.2006.01.06: modified for public                                   
//                                                                         
//

//***************************************************************************//
#include "Example.hpp"          // corresponding header file
#include <math.h>               // for atan2, sqrt
#include <stdio.h>              // for sample output
//***************************************************************************//

//***************************************************************************//
// plugin information
// Change this information to siut your needs. rFactor will use this information
// to ensure that it is using the most recent version
unsigned g_uPluginID          = 2;
char     g_szPluginName[]     = "ExampleInternalsPlugin - 2006.01.06";
unsigned g_uPluginVersion     = 002;
unsigned g_uPluginObjectCount = 1;
InternalsPluginInfo g_PluginInfo;
//***************************************************************************//

//***************************************************************************//
// interface to plugin information
extern "C" __declspec(dllexport)
const char* __cdecl GetPluginName() { return g_szPluginName; }

extern "C" __declspec(dllexport)
unsigned __cdecl GetPluginVersion() { return g_uPluginVersion; }

extern "C" __declspec(dllexport)
unsigned __cdecl GetPluginObjectCount() { return g_uPluginObjectCount; }

// get the plugin-info object used to create the plugin.
extern "C" __declspec(dllexport)
PluginObjectInfo* __cdecl GetPluginObjectInfo( const unsigned uIndex )
{
  switch(uIndex)
  {
    case 0:
      return  &g_PluginInfo;
    default:
      return 0;
  }
}
//***************************************************************************//

//***************************************************************************//
// InternalsPluginInfo class
InternalsPluginInfo::InternalsPluginInfo()
{
  // put together a name for this plugin
  sprintf( m_szFullName, "%s - %s", g_szPluginName, InternalsPluginInfo::GetName() );
}
//***************************************************************************//

//***************************************************************************//
const char*    InternalsPluginInfo::GetName()     const { return ExampleInternalsPlugin::GetName(); }
const char*    InternalsPluginInfo::GetFullName() const { return m_szFullName; }
// Change this to suit your needs
const char*    InternalsPluginInfo::GetDesc()     const { return "Example Internals Plugin"; }
const unsigned InternalsPluginInfo::GetType()     const { return ExampleInternalsPlugin::GetType(); }
const char*    InternalsPluginInfo::GetSubType()  const { return ExampleInternalsPlugin::GetSubType(); }
const unsigned InternalsPluginInfo::GetVersion()  const { return ExampleInternalsPlugin::GetVersion(); }
void*          InternalsPluginInfo::Create()      const { return new ExampleInternalsPlugin(); }
//***************************************************************************//

//***************************************************************************//
// InternalsPlugin class
const char ExampleInternalsPlugin::m_szName[] = "InternalsPlugin";
const char ExampleInternalsPlugin::m_szSubType[] = "Internals";
const unsigned ExampleInternalsPlugin::m_uID = 1;
const unsigned ExampleInternalsPlugin::m_uVersion = 1;
//***************************************************************************//

//***************************************************************************//
PluginObjectInfo *ExampleInternalsPlugin::GetInfo() { return &g_PluginInfo; }
//***************************************************************************//

//***************************************************************************//
void ExampleInternalsPlugin::Startup()
{
  // Open ports, read configs, whatever you need to do.  For now, I'll just clear out the
  // example output data.
  FILE *fo = fopen( "ExampleInternalsOutput.txt", "w" );
  fclose( fo );

  // default enabled to true
  mEnabled = true;
}
//***************************************************************************//

//***************************************************************************//
void ExampleInternalsPlugin::EnterRealtime()
{
  // start up timer every time we enter realtime
  mET = 0.0f;
}
//***************************************************************************//

//***************************************************************************//
void ExampleInternalsPlugin::UpdateTelemetry( const TelemInfo &info )
{
  const unsigned short cusBufferSize = 1024;
  char* cpTemp = new char[cusBufferSize];

  // Use the incoming data, for now I'll just write some of it to a file to a) make sure it
  // is working, and b) explain the coordinate system a little bit (see header for more info)
  FILE *fo = fopen( "ExampleInternalsOutput.txt", "a" );
  if( fo != NULL )
  {
    // Delta time is variable, as we send out the info once per frame
    sprintf( cpTemp, "DT=%.4f\n", info.mDeltaTime );
    fprintf( fo, "%s", cpTemp);
    mpConsole->Write(cpTemp);
    sprintf( cpTemp, "Lap=%d StartET=%.3f\n", info.mLapNumber, info.mLapStartET );
    fprintf( fo, "%s", cpTemp);
    mpConsole->Write(cpTemp);
    sprintf( cpTemp, "Vehicle=%s\n", info.mVehicleName );
    fprintf( fo, "%s", cpTemp);
    mpConsole->Write(cpTemp);
    sprintf( cpTemp, "Track=%s\n", info.mTrackName );
    fprintf( fo, "%s", cpTemp);
    mpConsole->Write(cpTemp);
    sprintf( cpTemp, "Pos=(%.3f,%.3f,%.3f)\n", info.mPos.x, info.mPos.y, info.mPos.z );
    fprintf( fo, "%s", cpTemp);
    mpConsole->Write(cpTemp);

    // Forward is roughly in the -z direction (although current pitch of car may cause some y-direction velocity)
    sprintf( cpTemp, "LocalVel=(%.2f,%.2f,%.2f)\n", info.mLocalVel.x, info.mLocalVel.y, info.mLocalVel.z );
    fprintf( fo, "%s", cpTemp);
    mpConsole->Write(cpTemp);
    sprintf( cpTemp, "LocalAccel=(%.1f,%.1f,%.1f)\n", info.mLocalAccel.x, info.mLocalAccel.y, info.mLocalAccel.z );
    fprintf( fo, "%s", cpTemp);
    mpConsole->Write(cpTemp);

    // Orientation matrix is left-handed
    sprintf( cpTemp, "[%6.3f,%6.3f,%6.3f]\n", info.mOriX.x, info.mOriX.y, info.mOriX.z );
    fprintf( fo, "%s", cpTemp);
    mpConsole->Write(cpTemp);
    sprintf( cpTemp, "[%6.3f,%6.3f,%6.3f]\n", info.mOriY.x, info.mOriY.y, info.mOriY.z );
    fprintf( fo, "%s", cpTemp);
    mpConsole->Write(cpTemp);
    sprintf( cpTemp, "[%6.3f,%6.3f,%6.3f]\n", info.mOriZ.x, info.mOriZ.y, info.mOriZ.z );
    fprintf( fo, "%s", cpTemp);
    mpConsole->Write(cpTemp);
    sprintf( cpTemp, "LocalRot=(%.3f,%.3f,%.3f)\n", info.mLocalRot.x, info.mLocalRot.y, info.mLocalRot.z );
    fprintf( fo, "%s", cpTemp);
    mpConsole->Write(cpTemp);
    sprintf( cpTemp, "LocalRotAccel=(%.2f,%.2f,%.2f)\n", info.mLocalRotAccel.x, info.mLocalRotAccel.y, info.mLocalRotAccel.z );
    fprintf( fo, "%s", cpTemp);
    mpConsole->Write(cpTemp);

    // Vehicle status
    sprintf( cpTemp, "Gear=%d RPM=%.1f\n", info.mGear, info.mEngineRPM );
    fprintf( fo, "%s", cpTemp);
    mpConsole->Write(cpTemp);
    sprintf( cpTemp, "Water=%.1f Oil=%.1f\n", info.mEngineWaterTemp, info.mEngineOilTemp );
    fprintf( fo, "%s", cpTemp);
    mpConsole->Write(cpTemp);
    sprintf( cpTemp, "ClutchRPM=%.1f\n", info.mClutchRPM );
    fprintf( fo, "%s", cpTemp);
    mpConsole->Write(cpTemp);

    // Driver input
    sprintf( cpTemp, "UnfilteredThrottle=%.1f%%\n", 100.0f * info.mUnfilteredThrottle );
    fprintf( fo, "%s", cpTemp);
    mpConsole->Write(cpTemp);
    sprintf( cpTemp, "UnfilteredBrake=%.1f%%\n", 100.0f * info.mUnfilteredBrake );
    fprintf( fo, "%s", cpTemp);
    mpConsole->Write(cpTemp);
    sprintf( cpTemp, "UnfilteredSteering=%.1f%%\n", 100.0f * info.mUnfilteredSteering );
    fprintf( fo, "%s", cpTemp);
    mpConsole->Write(cpTemp);
    sprintf( cpTemp, "UnfilteredClutch=%.1f%%\n", 100.0f * info.mUnfilteredClutch );
    fprintf( fo, "%s", cpTemp);
    mpConsole->Write(cpTemp);

    // Wheels
    sprintf( cpTemp, "SteeringArmForce=%.1f\n", info.mSteeringArmForce );
    fprintf( fo, "%s", cpTemp);
    mpConsole->Write(cpTemp);
    for( long i = 0; i < 4; ++i )
    {
      const TelemWheel &wheel = info.mWheel[i];
      sprintf( cpTemp, "Wheel=%s\n", (i==0)?"FrontLeft":(i==1)?"FrontRight":(i==2)?"RearLeft":"RearRight" );
      fprintf( fo, "%s", cpTemp);
      mpConsole->Write(cpTemp);
      sprintf( cpTemp, " ForwardRotation=%.1f\n", -wheel.mRotation );
      fprintf( fo, "%s", cpTemp);
      mpConsole->Write(cpTemp);
      sprintf( cpTemp, " SuspensionDeflection=%.3f RideHeight=%.3f\n", wheel.mSuspensionDeflection, wheel.mRideHeight );
      fprintf( fo, "%s", cpTemp);
      mpConsole->Write(cpTemp);
      sprintf( cpTemp, " TireLoad=%.1f LateralForce=%.1f GripFract=%.3f\n", wheel.mTireLoad, wheel.mLateralForce, wheel.mGripFract );
      fprintf( fo, "%s", cpTemp);
      mpConsole->Write(cpTemp);
      sprintf( cpTemp, " BrakeTemp=%.1f TirePressure=%.1f\n", wheel.mBrakeTemp, wheel.mPressure );
      fprintf( fo, "%s", cpTemp);
      mpConsole->Write(cpTemp);
      sprintf( cpTemp, " TireTemp(l/c/r)=%.1f/%.1f/%.1f\n", wheel.mTemperature[0], wheel.mTemperature[1], wheel.mTemperature[2] );
      fprintf( fo, "%s", cpTemp);
      mpConsole->Write(cpTemp);
    }

    // Compute some auxiliary info based on the above
    TelemVect3 forwardVector = { -info.mOriX.z, -info.mOriY.z, -info.mOriZ.z };
    TelemVect3    leftVector = {  info.mOriX.x,  info.mOriY.x,  info.mOriZ.z };

    // These are normalized vectors, and remember that our world Y coordinate is up.  So you can
    // determine the current pitch and roll (w.r.t. the world x-z plane) as follows:
    const float pitch = atan2f( forwardVector.y, sqrtf( ( forwardVector.x * forwardVector.x ) + ( forwardVector.z * forwardVector.z ) ) );
    const float  roll = atan2f(    leftVector.y, sqrtf( (    leftVector.x *    leftVector.x ) + (    leftVector.z *    leftVector.z ) ) );
    const float radsToDeg = 57.296f;
    sprintf( cpTemp, "Pitch = %.1f deg, Roll = %.1f deg\n", pitch * radsToDeg, roll * radsToDeg );
    fprintf( fo, "%s", cpTemp);
    mpConsole->Write(cpTemp);

    const float metersPerSec = sqrtf( ( info.mLocalVel.x * info.mLocalVel.x ) +
                                      ( info.mLocalVel.y * info.mLocalVel.y ) +
                                      ( info.mLocalVel.z * info.mLocalVel.z ) );
    sprintf( cpTemp, "Speed = %.1f KPH, %.1f MPH\n\n", metersPerSec * 3.6f, metersPerSec * 2.237f );
    fprintf( fo, "%s", cpTemp);
    mpConsole->Write(cpTemp);

    // Close file
    fclose( fo );
  }
  if(cpTemp)
    delete [] cpTemp;
}
//***************************************************************************//

//***************************************************************************//
void ExampleInternalsPlugin::UpdateGraphics( const GraphicsInfo &info )
{

  const unsigned short cusBufferSize = 1024;
  char* cpTemp = new char[cusBufferSize];

  // Use the incoming data, for now I'll just write some of it to a file to a) make sure it
  // is working, and b) explain the coordinate system a little bit (see header for more info)
  FILE *fo = fopen( "ExampleInternalsOutput.txt", "a" );
  if( fo != NULL )
  {
    // Print stuff
    sprintf( cpTemp, "CamPos=(%.1f,%.1f,%.1f)\n", info.mCamPos.x, info.mCamPos.y, info.mCamPos.z );
    fprintf( fo, "%s", cpTemp);
    mpConsole->Write(cpTemp);

    sprintf( cpTemp, "CamOri=(%.1f,%.1f,%.1f)\n", info.mCamOri.x, info.mCamOri.y, info.mCamOri.z );
    fprintf( fo, "%s", cpTemp);
    mpConsole->Write(cpTemp);

    sprintf( cpTemp, "HWND=%d\n\n", info.mHWND );
    fprintf( fo, "%s", cpTemp);
    mpConsole->Write(cpTemp);


    // Close file
    fclose( fo );
  }

  if(cpTemp)
    delete [] cpTemp;
}
//***************************************************************************//

//***************************************************************************//
bool ExampleInternalsPlugin::CheckHWControl( const char * const controlName, float &fRetVal )
{
  // only if enabled, of course
  if( !mEnabled )
    return( false );

  // Note that incoming value is the game's computation, in case you're interested.

  // Sorry, no control allowed over actual vehicle inputs ... would be too easy to cheat!
  // However, you can still look at the values.

  // Note: since the game calls this function every frame for every available control, you might consider
  // doing a binary search if you are checking more than 7 or 8 strings, just to keep the speed up.
  if( stricmp( controlName, "LookLeft" ) == 0 )
  {
    const float headSwitcheroo = fmodf( mET, 2.0f );
    if( headSwitcheroo < 0.5 )
      fRetVal = 1.0f;
    else
      fRetVal = 0.0f;
    return( true );
  }
  else if( stricmp( controlName, "LookRight" ) == 0 )
  {
    const float headSwitcheroo = fmodf( mET, 2.0f );
    if( ( headSwitcheroo > 1.0f ) && ( headSwitcheroo < 1.5f ) )
      fRetVal = 1.0f;
    else
      fRetVal = 0.0f;
    return( true );
  }

  return( false );
}
//***************************************************************************//

//***************************************************************************//
bool ExampleInternalsPlugin::ForceFeedback( float &forceValue )
{
  // Note that incoming value is the game's computation, in case you're interested.

  // CHANGE TO ENABLE FORCE EXAMPLE
  return( false );

  // I think the bounds are -11500 to 11500 ...
//  forceValue = 11500.0f * sinf( mET );
//  return( true );
}
//***************************************************************************//



